/* ppcrack 1.0 - decrypts PowerPacker packed / encrypted files
 * by Stuart Caie <kyzer@4u.net>, this software is in the Public Domain
 *
 * Webpage: http://www.kyzer.me.uk/pack/#ppcrack
 *
 * Usage: ppcrack <files>
 *        ppcrack -pass <key> <files>
 *        ppcrack -key <key> <files>
 *        ppcrack -from <key> -to <key> <files>
 *
 * Examples: ppcrack -key 034b5769 file.pp
 *           ppcrack -from 00000000 -to 1fffffff file.pp
 *           ppcrack -pass MyPassword *.pp
 *           ppcrack *.pp
 *
 * Anything which decrypts then decrunches to valid data is saved to disk
 * as <original filename>.<decryption key>
 *
 * As a bonus, if any file is a PowerPacked, but not encrypted,
 * it will be decrunched anyway, and saved as <original filename>.decrunched
 */

/* version history
 * 0.1  09-Dec-2001 - original release
 * 0.2  25-Nov-2002 - fixed off-by-one error detection error in PP_BYTE_OUT()
 *                  - allowed decrunching to end after a literal run.
 *                  - now ppValidate doesn't need the buffer decrypted.
 *                  - new command-line interface
 * 0.3  07-Sep-2003 - vastly improved skipping of known incorrect keys
 *                  - display update now includes keys/second
 *                  - improved documentation
 * 0.4  18-Sep-2003 - smaller command line parser
 *                  - support for PPLS files
 *                  - now writes key update / fatal errors to stderr
 *                  - split the basic PowerPacker functions into pplib.c
 *                  - support large files that aren't fully encrypted (!)
 *                  - improved documenatation
 * 0.4b 18-Sep-2003 - added "-big" option; the big files problem is only
 *                    in some unofficial versions of powerpacker.library.
 * 1.0  21-Nov-2010 - unpacks executables
 *                  - dictionary attack on encrypted executables
 *                  - "-cpus" option for multithreaded key search
 *                  - stops after first success, new "-all" option to try all
 * 1.0a 25-Nov-2010 - made compatible with C89 compilers
 */

#ifdef NO_PTHREADS
# define pthread_cancel(a)
# define pthread_join(a,b)
# define pthread_create(a,b,c,d) 0
# define pthread_exit(a) exit(1)
typedef unsigned int pthread_t;
#else
# include <pthread.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include <pplib.h>

/* byte-order macros */
#define EndGetM32(a) ((((a)[0])<<24)|(((a)[1])<<16)|(((a)[2])<<8)|((a)[3]))
#define EndGetM16(a) ((((a)[0])<<8)|((a)[1]))
#define EndPutM32(a,val) do {     \
    unsigned int dval = (val);    \
    (a)[0] = (dval >> 24) & 0xFF; \
    (a)[1] = (dval >> 16) & 0xFF; \
    (a)[2] = (dval >>  8) & 0xFF; \
    (a)[3] = (dval)       & 0xFF; \
} while (0)

/* pointer alignment macros */
#ifdef __cplusplus
# define WORD_ALIGN(t, v)     (v=(t)((size_t(v) + 1) & ~1))
# define LONGWORD_ALIGN(t, v) (v=(t)((size_t(v) + 3) & ~3))
#else
# define WORD_ALIGN(t, v)     (v=(t)((((size_t)v) + 1) & ~1))
# define LONGWORD_ALIGN(t, v) (v=(t)((((size_t)v) + 3) & ~3))
#endif

/* logging macros */
#define WARN(msg) fprintf(stderr, "%s: %s\n", pp->name, msg)
#define FAIL(msg) do {WARN(msg); return 0;} while (0)

/* Amiga executable hunk types */
#define HUNK_CODE    0x3E9
#define HUNK_DATA    0x3EA
#define HUNK_BSS     0x3EB
#define HUNK_RELOC32 0x3EC
#define HUNK_HEADER  0x3F3
#define HUNK_END     0x3F2
#define HUNK_OVERLAY 0x3F5
#define HUNK_BREAK   0x3F6

/* ppfile flags */
#define EXECUTABLE   (1<<0) /* file is an executable                    */
#define ENCRYPTED    (1<<1) /* file is encrypted                        */
#define OVERLAY      (1<<2) /* file has an overlay section to copy      */
#define MASTERMODE   (1<<3) /* 2.x / 3.x executable in master mode      */
#define BIGMODE      (1<<4) /* Kev Harrison's broken library was used   */
#define PP40         (1<<5) /* PP4.0 executable, needs reloc correction */
#define PPLS         (1<<6) /* PPLS file: PP4.0 relocs, maybe overlay   */
#define ALLRESULTS   (1<<7) /* continue after finding a good key        */

struct ppfile {
    const char *name;             /* name of the file */
    const unsigned char *buf;     /* buffer containing the entire file */
    unsigned int buf_len;         /* length of the buffer */
    unsigned int src_off;         /* offset to the packed data */
    unsigned int efficiency_off;  /* offset to the "efficiency" bytes */
    unsigned int src_len;         /* length of the packed data */
    unsigned int encrypted_len;   /* amount of the encrypted data */
    unsigned int dest_len;        /* length of the unpacked data */
    unsigned int max_dest_len;    /* maximum unpacked length (PP40|PPLS) */
    unsigned int overlay_off;     /* offset to the PPLS overlay section */
    unsigned int flags;
};

/*--- final decrunching ------------------------------------------------------*/

/* Executables compressed with any variant of PowerPacker 4.0 have their
 * RELOC32 hunk extensions packed with delta compression.
 *
 * NORMAL RELOC32 FORMAT
 * 4 bytes: number of relocs (n)
 * 4 bytes: which hunk they point to
 * 4*n bytes: list of offsets to relocate
 * -- repeat all the above until n==0
 *
 * POWERPACKED RELOC32 FORMAT
 * 2 bytes: number of relocs (n)
 * 2 bytes: which hunk they point to
 * 4 bytes: first reloc offset
 * (4 or 1 bytes) * n-1:
 *   if first byte is zero, then 4 byte increment of offset
 *   if first byte is nonzero, then 1 byte increment to offset
 * -- repeat all the above until n==0
 */
static int unpack_relocs(struct ppfile *pp, unsigned char *dest) {
    unsigned int num_hunks, hunk_len, extra_len, unpacked_len;
    unsigned int num_relocs, offset, delta;
    unsigned char *p, *rp, *end;

    num_hunks = EndGetM32(&dest[0x10]) - EndGetM32(&dest[0x0C]) + 1;
    p   = &dest[0x14 + (num_hunks * 4)];
    end = &dest[pp->dest_len];

    while (p < end) {
	unsigned int hunk_type = EndGetM32(p) & 0x1FFFFFFF;
	p += 4;

	switch (hunk_type) {
	case HUNK_CODE:
	case HUNK_DATA:
	    hunk_len = EndGetM32(p) & 0x3FFFFFFF;
	    p += 4 + (hunk_len * 4);
	    num_hunks--;
	    break;

	case HUNK_BSS:
	    p += 4; /* don't need to read the size, just skip it */
	    num_hunks--;
	    break;
	    
	case HUNK_END:
	    /* nothing needed */
	    break;
	    
	case HUNK_RELOC32:
	    rp = p;

	    /* calculate the unpacked size of the relocs */
	    unpacked_len = 0;
	    while ((num_relocs = EndGetM16(rp)) > 0) {
		unpacked_len += 8 + (4 * num_relocs), rp += 8;
		while (--num_relocs > 0) rp += (*rp) ? 1 : 4;
		WORD_ALIGN(unsigned char *, rp);
	    }
	    unpacked_len += 4, rp += 2;
	    LONGWORD_ALIGN(unsigned char *, rp);

	    /* if no relocs are found, it's likely that the hunk
	     * is not packed - skip through it to the end */
	    if (unpacked_len == 4) {
		while ((num_relocs = EndGetM32(p)) && p < end) {
		    p += 8 + (4 * num_relocs);
		}
		p += 4;
		break;
	    }

	    /* expand the space used by the relocs from packed length to
	     * unpacked length; move the rest of the executable up */
	    extra_len = unpacked_len - (rp - p);
	    if (pp->dest_len + extra_len > pp->max_dest_len) {
		FAIL("PP4.0 relocs expanded too far; file is corrupt");
	    }
	    rp = p + extra_len;
	    memmove(rp, p, end - p);
	    end += extra_len;
	    pp->dest_len += extra_len;

	    /* unpack the relocs */
	    while ((num_relocs = EndGetM16(rp)) > 0) {
		EndPutM32(p, num_relocs); /* number of relocs */
		EndPutM32(p+4, EndGetM16(rp+2)); /* hunk number */
		offset = EndGetM32(rp+4);
		p += 8, rp += 8;
		while (--num_relocs) {
		    if ((delta = *rp++) == 0) {
			delta = (rp[0] << 16) | (rp[1] << 8) | rp[2];
			rp += 3;
		    }
		    EndPutM32(p, offset); p += 4;
		    offset += delta * 2;
		}
		EndPutM32(p, offset); p += 4;
		WORD_ALIGN(unsigned char *, rp);
	    }
	    /* store final number of relocs (0) */
	    EndPutM32(p, 0); p += 4;
	    break;
	    
	default:
	    /* unknown hunk type - abort (but succeed) */
	    WARN("unknown hunk type; file may be corrupt");
	    return 1;
	}
    }

    /* success? */
    if (num_hunks != 0) FAIL("failed to fix relocs; file is corrupt");
    return 1;
}

static int decrunch(struct ppfile *pp, unsigned int key) {
    const unsigned char *src = &pp->buf[pp->src_off];
    char *outname = (char *) malloc(strlen(pp->name) + 12);
    unsigned char *dest = (unsigned char *) malloc(
	((pp->flags & (PP40|PPLS)) ? pp->max_dest_len : pp->dest_len) +
	((pp->flags & OVERLAY) ? pp->buf_len - pp->src_len : 0));
    unsigned char *decsrc = (unsigned char *)
	((pp->flags & ENCRYPTED) ? malloc(pp->src_len + 4) : NULL);
    FILE *fh = NULL;
    int result = 0;

    /* did we allocate all required memory? */
    if (!outname || !dest || ((pp->flags & ENCRYPTED) && !decsrc)) {
	WARN("out of memory!");
	goto cleanup;
    }

    /* generate output filename */
    snprintf(outname, strlen(pp->name) + 12,
	     (pp->flags & ENCRYPTED) ? "%s.%08X" : "%s.decrunched",
	     pp->name, key);

    if (pp->flags & ENCRYPTED) {
	/* copy src to decsrc so we can modify it */
	memcpy(decsrc, src, pp->src_len + 4);

	/* decrypt decsrc with this key */
	ppDecrypt(decsrc, pp->encrypted_len, key);

	/* use this as the source data */
	src = decsrc;
    }

    /* decrunch buffer as appropriate */
    result = (pp->flags & MASTERMODE)
	? ppDecrunchBuffer_m(&pp->buf[pp->efficiency_off], src, dest,
			     pp->src_len, pp->dest_len)
        : ppDecrunchBuffer  (&pp->buf[pp->efficiency_off], src, dest,
			     pp->src_len, pp->dest_len);
    if (!result) {
	WARN("failed to decrunch");
	goto cleanup;
    }

    /* PowerPacker 4.x executables and PPLS files have packed RELOC32 hunks */
    if (pp->flags & (PP40 | PPLS)) {
	if (!unpack_relocs(pp, dest)) {
	    result = 0;
	    goto cleanup;
	}

	/* PPLS files can have an overlay (included in max_dest_len); copy it */
	if ((pp->flags & PPLS) && pp->overlay_off) {
	    unsigned int overlay_len = pp->buf_len - pp->overlay_off;
	    unsigned int copy_len = pp->max_dest_len - pp->dest_len;
	    if (overlay_len < copy_len) copy_len = overlay_len;
	    memcpy(&dest[pp->dest_len], &pp->buf[pp->overlay_off], copy_len);
	    pp->dest_len += copy_len;
	}
    }

    if (pp->flags & OVERLAY) {
	/* start of overlay hunk (points at HUNK_OVERLAY) */
	unsigned int ovl_off = pp->src_off + pp->src_len + 8;
	unsigned int ovl_len = pp->buf_len - ovl_off, i;
	unsigned char *p, *overlay = &dest[pp->dest_len];

	/* copy overlay from src to end of dest */
	memcpy(overlay, &pp->buf[ovl_off], ovl_len);

	/* The offsets inside the overlay hunk need to be moved from the
	 * right position for a packed file to the right position for an
	 * unpacked file.
	 *
	 * pp->dest_len = start of overlay hunk in unpacked executable
	 * ovl_off      = start of overlay hunk in packed executable
	 */
	for (i=EndGetM32(&overlay[8]), p=&overlay[12]; i--; p += 8) {
	    EndPutM32(p, EndGetM32(p) + (pp->dest_len - ovl_off));
	}
	pp->dest_len += ovl_len;
    }

    /* save the decrunched buffer */
    if ((fh = fopen(outname, "wb")) &&
	(fwrite(dest, 1, pp->dest_len, fh) == pp->dest_len) &&
	(fclose(fh) == 0))
    {
	fh = NULL;
	printf("%s: decrunched to %s\n", pp->name, outname);
    }
    else {
	perror(pp->name);
	WARN("decrunched OK but failed to save output file");
	result = 0;
    }

 cleanup:
    if (fh) fclose(fh);
    free(outname);
    free(dest);
    free(decsrc);
    return result;
}

/*--- finding correct decryption key -----------------------------------------*/

struct pp_unit {
    struct pp_units *base;
    unsigned int start_key, end_key, current_key, success;
};

struct pp_partial_state {
    const unsigned char *buf, *olens;
    unsigned char state, bits_todo, lit_bit, skip_bits;
    unsigned int x, y, offset, written, dest_len;
};

struct pp_units {
    unsigned int num_units;
    struct ppfile *pp;
    struct {
	unsigned int drop_key, drop_mask;
    } fast;
    struct {
	struct pp_partial_state state;
	unsigned int num_longs;
    } partial;
    pthread_t *thread;
    struct pp_unit *unit;
};

#define UPDATE_PROGRESS							\
    if ((keys_todo & time_mask) == 0) {					\
	unsigned int i, keys_done = 0, keys_total = 0;			\
	/* regulate time_mask to be roughly once a second */		\
	if ((time_now = time(NULL)) == time_then) {			\
	    time_mask = (time_mask << 1) | 1;				\
	}								\
	time_then = time_now;						\
	/* write our current progress */				\
	unit->current_key = key;					\
	/* count up all current progress */				\
	for (i = 0; i < unit->base->num_units; i++) {			\
	    keys_done  += unit->base->unit[i].current_key		\
		- unit->base->unit[i].start_key;			\
	    keys_total += unit->base->unit[i].end_key + 1		\
		- unit->base->unit[i].start_key;			\
	    /* if anyone else has succeeded, exit now */		\
	    if (unit->base->unit[i].success &&				\
		!(unit->base->pp->flags & ALLRESULTS))			\
	    {								\
		/* cancel everyone */					\
	        for (i = 0; i < unit->base->num_units-1; i++) {		\
		    /* except ourselves */				\
		    if (unit->base->unit[i].current_key != key) {	\
			pthread_cancel(unit->base->thread[i]);		\
		    }							\
		}							\
		return NULL;						\
	    }								\
	}								\
	/* avoid dividing by zero due to overflow */			\
	if (keys_total == 0) keys_total = 0xFFFFFFFF;			\
	/* print current progress */					\
	printf("%-2.02f%% complete    \r",				\
	       (100.0 * (double)keys_done) / (double)keys_total);	\
	fflush(stdout);							\
    }


#define READ_BITS(nb, v) do {					\
    for (bc = (nb); bl < bc; bl += 8) {				\
	if (buf < src) goto done; /* out of source bits */	\
	bb |= ((*--buf ^ *--kp) << bl); if (kp==ka) kp=&ka[4];	\
    }								\
    (v)=0; bl-=bc; while(bc--){(v)<<=1; (v)|=bb&1; bb>>=1;}	\
} while (0)

#define PPCRACK_FAST(masterbit)						\
    struct pp_unit *unit = (struct pp_unit *) p;			\
    unsigned int key           = unit->start_key;			\
    unsigned int keys_todo     = unit->end_key - key;			\
    unsigned int drop_key      = unit->base->fast.drop_key;		\
    unsigned int drop_mask     = unit->base->fast.drop_mask;		\
    const unsigned char *buf   = unit->base->pp->buf;			\
    const unsigned char *src   = &buf[unit->base->pp->src_off];		\
    const unsigned char *olens = &buf[unit->base->pp->efficiency_off];	\
    unsigned int src_len       = unit->base->pp->src_len;		\
    unsigned int dest_len      = unit->base->pp->dest_len;		\
    unsigned int time_mask     = 1;					\
    time_t time_then = time(NULL), time_now;				\
    unsigned char ka[4], *kp, skip_bits = src[src_len+3];		\
    unsigned int written, bb, x, y, ob, offset, bl, bc;			\
    do {								\
	if ((key & drop_mask) == drop_key) {				\
	    key += drop_mask;						\
	    if (keys_todo < drop_mask) break;				\
	    keys_todo -= drop_mask;					\
	    continue;							\
	}								\
	UPDATE_PROGRESS;						\
	ka[0] = key >> 24;						\
	ka[1] = key >> 16;						\
	ka[2] = key >> 8;						\
	ka[3] = key;							\
	written = 0; buf = &src[src_len]; kp = &ka[4]; bl = 0; bb = 0;	\
	READ_BITS(skip_bits, x);					\
	while (written < dest_len) {					\
	    READ_BITS(1, x);						\
	    if (x == masterbit) {					\
		y = 1; do { READ_BITS(2, x); y += x; } while (x == 3);	\
		written += y; if (written > dest_len) break;		\
		while (y--) READ_BITS(8, x);				\
		if (written == dest_len) break;				\
	    }								\
	    READ_BITS(2, x); ob = olens[x]; y = x + 2;			\
	    if (x == 3) {						\
		READ_BITS(1, x); if (x==0) ob=7; READ_BITS(ob, offset);	\
		do { READ_BITS(3, x); y += x; } while (x == 7);		\
	    } else { READ_BITS(ob, offset); }				\
	    if (offset >= written) break;				\
	    written += y; if (written > dest_len) break;		\
	}								\
done:  if (written == dest_len) {					\
	    if (decrunch(unit->base->pp, key)) unit->success++;		\
	}								\
	key++;								\
    } while (keys_todo-- > 0);

static void *ppcrack_fast_normal(void *p) {
    PPCRACK_FAST(0);
    return NULL;
}
static void *ppcrack_fast_master(void *p) {
    PPCRACK_FAST(1);
    return NULL;
}

enum {SKIP_BITS, LIT_BIT, LIT_LEN, LITERALS, MATCH_LEN,
      OFFSET_AND_LEN, OAL_OFFSET, OAL_ADDLEN, OFFSET, MATCH_DONE};

static void partial_init(struct pp_partial_state *pps, struct ppfile *pp,
			 int lit_bit)
{
    pps->buf       = &pp->buf[pp->src_off + pp->src_len];
    pps->olens     = &pp->buf[pp->efficiency_off];
    pps->state     = SKIP_BITS;
    pps->lit_bit   = lit_bit;
    pps->skip_bits = pps->buf[3];
    pps->written   = 0;
    pps->dest_len  = pp->dest_len;
}

#define DUMP_BITS(n) bb >>= n, bl -= n
#define READ_BITS_PARTIAL(var)			\
    while (bl && bits_todo) {			\
	var = (var << 1) | (bb & 1);		\
	DUMP_BITS(1);				\
	bits_todo--;				\
    }

static int partial_verify(struct pp_partial_state *pps,
			  unsigned int num_longs,
			  unsigned int key)
{
    const unsigned char *buf = pps->buf;
    unsigned int state       = pps->state;
    unsigned int bits_todo   = pps->bits_todo;
    unsigned int lit_bit     = pps->lit_bit;
    unsigned int skip_bits   = pps->skip_bits;
    unsigned int x           = pps->x;
    unsigned int y           = pps->y;
    unsigned int offset      = pps->offset;

    while (num_longs--) {
	unsigned int bb, bl = 32;
	buf -= 4; bb = EndGetM32(buf) ^ key;
	while (bl > 0) {
	    switch (state) {
	    case SKIP_BITS:
		DUMP_BITS(skip_bits);
		/* fall through to LIT_BIT */

	    case LIT_BIT:
		state = ((bb & 1) == lit_bit) ? LIT_LEN : MATCH_LEN;
		DUMP_BITS(1); y=1, x=0, bits_todo=2;
		break;

	    case LIT_LEN:
		READ_BITS_PARTIAL(x);
		if (bits_todo == 0) {
		    y += x;
		    if (x == 3) x=0, bits_todo=2;
		    else pps->written += y, state=LITERALS, bits_todo=8*y;
		}
		break;

	    case LITERALS:
		x = (bl < bits_todo) ? bl : bits_todo;
		DUMP_BITS(x); bits_todo -= x;
		if (bits_todo == 0) state=MATCH_LEN, bits_todo=2, x=0;
		break;

	    case MATCH_LEN:
		READ_BITS_PARTIAL(x);
		if (bits_todo == 0) {
		    y = x + 2;
		    if (x == 3) state=OFFSET_AND_LEN, x=pps->olens[x];
		    else        state=OFFSET, bits_todo=pps->olens[x], offset=0;
		}
		break;

	    case OFFSET_AND_LEN:
		bits_todo = (bb & 1) ? x : 7;
		DUMP_BITS(1); state=OAL_OFFSET, offset=0;
		break;

	    case OAL_OFFSET:
		READ_BITS_PARTIAL(offset);
		if (bits_todo == 0) state=OAL_ADDLEN, bits_todo=3, x=0;
		break;

	    case OAL_ADDLEN:
		READ_BITS_PARTIAL(x);
		if (bits_todo == 0) {
		    y += x;
		    if (x == 7) bits_todo=3, x=0; else state=MATCH_DONE;
		}
		break;

	    case OFFSET:
		READ_BITS_PARTIAL(offset);
		if (bits_todo == 0) {
		    if (offset >= pps->written) return 0;
		    state = MATCH_DONE;
		}
		break;

	    case MATCH_DONE:
		pps->written += y;
		if (pps->written > pps->dest_len) return 0;
		state = LIT_BIT;
		break;
	    } /* switch (state) */
	} /* while (bl > 0) */
    } /* while (num_longs--) */

    /* save state */
    pps->buf       = buf;
    pps->state     = state;
    pps->bits_todo = bits_todo;
    pps->x         = x;
    pps->y         = y;
    pps->offset    = offset;
    return 1;
}

static void *ppcrack_partial(void *p) {
    struct pp_unit *unit = (struct pp_unit *) p;
    unsigned int key            = unit->start_key;
    unsigned int keys_todo      = unit->end_key - key;
    struct pp_partial_state pps = unit->base->partial.state;
    unsigned int num_longs = unit->base->partial.num_longs;
    unsigned int time_mask = 1;
    time_t time_then = time(NULL), time_now;

    do {
	UPDATE_PROGRESS;
        if (partial_verify(&pps, num_longs, key)) {
	    if (pps.written == pps.dest_len) {
		if (decrunch(unit->base->pp, key)) unit->success++;
	    }
	    pps = unit->base->partial.state;
        }
	key++;
    } while (keys_todo-- > 0);
    return NULL;
}

/*--- recognition of executables ---------------------------------------------*/

struct pp_exe_version {
    const char *name;

    struct {
	/* if offset is nonzero, then there is a master-mode
	 * if buf[offset].w == master, then master-mode is on
	 * if buf[offset].w == normal, then master-mode is off */
	unsigned short offset, master, normal;
    } master;

    /* two longword tests */
    struct {
	unsigned short offset;
	unsigned int   val;
    }  t1, t2;

    unsigned short min_len;        /* smallest valid executable length */
    unsigned short efficiency_off; /* offset to the bit lengths (efficiency) */
    unsigned short src_len_off;    /* offset to the length of the data */
    unsigned short cksum_off;      /* offset to the password checksum (.w) */
};

/* all tests and offsets taken from the XFD PowerPacker slave:
 * http://libxad.cvs.sf.net/viewvc/libxad/libxfd/source/master/slaves/PowerPacker.a
 */
static const struct pp_exe_version pp_versions[] = {
    /*name            master                 t1                 t2                 minlen effoff srclen cksum */
    {"2.0",           {0x1A0,0x6720,0x6620}, {0x44,0x45FA0138}, {0x4C,0x61000134}, 0x240, 0x228, 0x17E, 0},
    {"2.3",           {0x19E,0x6720,0x6620}, {0x44,0x45FA0136}, {0x4C,0x61000132}, 0x240, 0x228, 0x17C, 0},
    {"3.0",           {0x1C0,0x643E,0x653E}, {0x24,0x487A0178}, {0x48,0x61000156}, 0x25C, 0x240, 0x44,  0},
    {"3.0 Encrypted", {0x23C,0x643E,0x653E}, {0x24,0x487A01F6}, {0xC6,0x61000156}, 0x318, 0x2FE, 0xC2,  0x8E},
    {"3.0 Overlay",   {0x1D6,0x643E,0x653E}, {0x3C,0x487A0176}, {0x60,0x61000154}, 0x27C, 0x256, 0x5C,  0},
    {"4.0",                         {0,0,0}, {0x24,0x487A01C8}, {0x48,0x610001A6}, 0x2B0, 0x290, 0x44,  0},
    {"4.0 Library",                 {0,0,0}, {0x24,0x487A0058}, {0x50,0x4EAEFFB2}, 0xB4,  0x94,  0x44,  0},
    {"4.0 Overlay",                 {0,0,0}, {0x3C,0x487A01C8}, {0x60,0x610001A6}, 0x2D4, 0x2A8, 0x5C,  0},
    {"4.0 Library Overlay",         {0,0,0}, {0x3C,0x487A0058}, {0x68,0x4EAEFFA6}, 0xD8,  0xAC,  0x5C,  0},
    {"4.0 Encrypted",               {0,0,0}, {0x24,0x487A0246}, {0xC6,0x610001A6}, 0x36C, 0x34E, 0xC2,  0x8E},
    {"4.0 Library Encrypted",       {0,0,0}, {0x24,0x487A005C}, {0x54,0x4EAEFFAC}, 0xB8,  0x98,  0x44,  0x4A}
};

/* recognised bona-fide PowerPacker crunched executables */
static int recog_pp_exe(struct ppfile *pp) {
    unsigned int i;

    /* PowerPacked executable file
     * 0x00 HUNK_HEADER (0x3F3)
     * 0x04 version (0)
     * 0x08 number of hunks (2 for normal, 7 or 12 for overlay),
     * 0x0C first hunk in this section (0)
     * 0x10 last hunk in this section (1),
     * 0x14 table of hunk sizes: hunk 0 size in longwords
     * 0x18 table of hunk sizes: hunk 1 size in longwords
     * 0x1C hunk 0 header (HUNK_CODE=0x3E9)
     * 0x20 hunk 0 size in longwords (matches value at 0x14)
     * 0x24 hunk 0 data
     * ---- after 0x24+([data at 0x14] * 4) bytes:
     * .... optional HUNK_END (0x3F2)
     * .... hunk 1 header (HUNK_DATA=0x3EA)
     * .... hunk 1 size in longwords (matches data at 0x18)
     * .... hunk 1 data (powerpacked data)
     */
    if (pp->buf_len < 0xB4                       ||
	EndGetM32(&pp->buf[0x00]) != HUNK_HEADER ||
	EndGetM32(&pp->buf[0x04]) != 0           ||
	EndGetM32(&pp->buf[0x08]) < 2            ||
	EndGetM32(&pp->buf[0x0C]) != 0           ||
	EndGetM32(&pp->buf[0x10]) != 1           ||
	EndGetM32(&pp->buf[0x1C]) != HUNK_CODE   ||
	EndGetM32(&pp->buf[0x14]) != EndGetM32(&pp->buf[0x20]))
    {
	return 0;
    }

    /* get the offset of the data hunk */
    pp->src_off = (EndGetM32(&pp->buf[0x14]) << 2) + 0x24;
    if (EndGetM32(&pp->buf[pp->src_off]) == HUNK_END) pp->src_off += 4;
    if (EndGetM32(&pp->buf[pp->src_off]) != HUNK_DATA) {
	return 0;
    }
    /* go from the data hunk header to the powerpacked data */
    pp->src_off += 8;

    /* check against PP version recognition table */
    for (i = 0; i < (sizeof(pp_versions) / sizeof(*pp_versions)); i++) {
	const struct pp_exe_version *x = &pp_versions[i];

	/* does this executable pass both tests? */
	if (pp->buf_len < x->min_len ||
	    (unsigned int)EndGetM32(&pp->buf[x->t1.offset]) != x->t1.val ||
	    (unsigned int)EndGetM32(&pp->buf[x->t2.offset]) != x->t2.val)
	{
	    continue;
	}

	/* test for master mode if present */
	if (x->master.offset) {
	    unsigned short val = EndGetM16(&pp->buf[x->master.offset]);
	    if      (val == x->master.master) pp->flags |= MASTERMODE;
	    else if (val != x->master.normal) continue;
	}

	/* this is the right type - copy parameters */
	pp->efficiency_off = x->efficiency_off;
	pp->src_len = EndGetM32(&pp->buf[x->src_len_off]);

	if (x->cksum_off) {
	    printf("%s: PowerPacker %s executable (sum=%04X)\n",
		   pp->name, x->name, EndGetM16(&pp->buf[x->cksum_off]));

	    /* Encrypted executables may have only part of the data encrypted.
	     * For PP3.0/PP4.0, the count for a DBRA loop decrypting longwords
	     * is stored at 0xB2 in the executable. For PP4.0 Lib, all bytes
	     * except the final longword are encrypted.
	     */
	    pp->encrypted_len = (x->cksum_off == 0x8E)
		? (EndGetM16(&pp->buf[0xB2]) + 1) * 4
		: pp->src_len - 4;
	    pp->flags |= ENCRYPTED;
	}
	else {
	    printf("%s: PowerPacker %s executable\n", pp->name, x->name);
	}

	/* add pp->flags according to name */
	if (x->name[0] == '4') {
	    pp->flags |= PP40;
	    pp->max_dest_len = EndGetM32(&pp->buf[pp->efficiency_off + 4]);
	}
	if (strchr(x->name, 'O')) {
	    pp->flags |= OVERLAY;
	}

	/* pp->src_len from the executable also counts the DDDDDDBB longword. */
	if (pp->src_len < 8) return 0;
	pp->src_len -= 4;

	/* success! */
	return 1;
    }

    /* fail if we went through the entire table without recognising anything */
    return 0;
}

/* Quick PowerPacker executables */
static int recog_qpp_exe(struct ppfile *pp) {
    if (pp->buf_len < 0x378) return 0;

    if (EndGetM32(&pp->buf[0x24]) == 0x487A01C4 &&
	EndGetM32(&pp->buf[0x42]) == 0x610001A8)
    {
	/* source length is a word (16 bits) */
	pp->src_len = EndGetM16(&pp->buf[0x40]);

	if (EndGetM16(&pp->buf[0x362]) == HUNK_DATA) {
	    pp->src_off        = 0x368;
	    pp->efficiency_off = 0x336;
	}
	else if (EndGetM16(&pp->buf[0x36E]) == HUNK_DATA) {
	    pp->src_off        = 0x374;
	    pp->efficiency_off = 0x340;
	}
	else return 0;
    }
    else if (EndGetM32(&pp->buf[0x24]) == 0x487A01C6 &&
	     EndGetM32(&pp->buf[0x44]) == 0x610001A8)
    {
	/* source length is a longword (32 bits) */
	pp->src_len = EndGetM16(&pp->buf[0x40]);

	if (EndGetM16(&pp->buf[0x366]) == HUNK_DATA) {
	    pp->src_off        = 0x36C;
	    pp->efficiency_off = 0x338;
	}
	else if (EndGetM16(&pp->buf[0x36E]) == HUNK_DATA) {
	    pp->src_off        = 0x374;
	    pp->efficiency_off = 0x342;
	}
	else return 0;

    }
    else return 0;

    printf("%s: Quick PowerPacker executable\n", pp->name);

    /* needs reloc expansion */
    pp->flags |= PP40;

    /* estimate max dest len based on dest_len */
    pp->max_dest_len = (EndGetM32(&pp->buf[pp->src_off+pp->src_len-4]) >> 8)*2;
    return 1;
}

struct ppp_type {
    unsigned int data_off, efficiency_off, src_len_off, src_off, max_dest_len;
};

static const struct ppp_type ppp_types[] = {
    /*data  eff    srcl  srcoff max_dest_len*/
    {0x2B4, 0x2A2, 0x44, 0x2BC, 0x2A6}, /* ppp0 */
    {0x2B0, 0x296, 0x44, 0x2B8, 0x29A}, /* ppp1 */
    {0x2AC, 0x292, 0x44, 0x2B4, 0x296}, /* ppp2 */
    {0x2EC, 0x292, 0x44, 0x2F4, 0x296}, /* ppp8,ppp9 */
    {0x2B4, 0x29A, 0xAC, 0x2BC, 0x29E}, /* ppp3,ppp4 */
    {0x2D8, 0x2B8, 0xAC, 0x2E0, 0x2BC}, /* ppp5  */
};

/* PowerPacker Patch executables */
static int recog_ppp_exe(struct ppfile *pp) {
    const struct ppp_type *pt;

    if (pp->buf_len < 0x278) return 0;

    if (EndGetM16(&pp->buf[0x24]) == 0x487A &&
	EndGetM16(&pp->buf[0x48]) == 0x6100)
    {
	unsigned int x = EndGetM16(&pp->buf[0x26]);
	unsigned int y = EndGetM16(&pp->buf[0x4A]);
	if      (x == 0x1DC && y == 0x1C0) pt = ppp_types+0;
	else if (x == 0x1DC && y == 0x1BA) pt = ppp_types+1;
	else if (x == 0x1E2 && y == 0x1C0) pt = ppp_types+2;
	else if (x == 0x1B6 && y == 0x1C0) pt = ppp_types+3;
	else return 0;
    }
    else if (EndGetM16(&pp->buf[0x24]) == 0x6066 &&
	     EndGetM32(&pp->buf[0x8C]) == 0x487A0156)
    {
	pt = ppp_types + ((EndGetM16(&pp->buf[0x22]) == 0x00AC) ? 5 : 4);
    }
    else return 0;

    if (EndGetM32(&pp->buf[pt->data_off]) == HUNK_DATA) {
	printf("%s: PowerPacker Patch executable\n", pp->name);
	pp->efficiency_off = pt->efficiency_off;
	pp->src_len        = EndGetM32(&pp->buf[pt->src_len_off]) - 4;
	pp->src_off        = pt->src_off;
	pp->max_dest_len   = EndGetM32(&pp->buf[pt->max_dest_len]);
	pp->flags         |= PP40;
	return 1;
    }
    return 0;
}

/*--- recognition of data files ----------------------------------------------*/

static int recog_pp_data(struct ppfile *pp) {
    unsigned int i;

    switch (EndGetM32(&pp->buf[0])) {
    /* standard data file (not encrypted)
     * 0x00 identifier 'PP20' or variant
     * 0x04 efficiency: 0x0w0x0y0z where w=9, x=9-10, y=9-12, z=9-13
     * 0x08 packed data, always multiple of 4 bytes
     * ---- 0xDDDDDDBB; DDDDDD=decrunched length, BB=start bit (0-31)
     */
    case 0x50503230: /* PP20 */
    case 0x5050324F: /* PP2O */
    case 0x5041434B: /* PACK */
    case 0x4D4C4443: /* MLDC */
    case 0x44454E21: /* DEN! */
    case 0x4D443132: /* MD12 */
    case 0x47415A21: /* GAZ! */
    case 0x58583530: /* XX50 */
    case 0x4C523838: /* LR88 */
    case 0x29075337:
	printf("%s: PowerPacker data file (%4.4s)\n", pp->name, pp->buf);
	pp->efficiency_off = 4;
	pp->src_off        = 8;
	break;

    /* PPLoadSeg file (not encrypted)
     * 0x00 identifier 'PPLS'
     * 0x04 decrunched length (includes reloc expansion + overlay)
     * 0x08 efficiency: 0x0w0x0y0z where w=9, x=9-10, y=9-12, z=9-13
     * 0x0C packed data, always multiple of 4 bytes
     * ---- 0xDDDDDDBB; DDDDDD=decrunched length, BB=start bit (0-31)
     */
    case 0x50504C53: /* PPLS */
	printf("%s: PPLoadSeg file\n", pp->name);
	pp->efficiency_off = 8;
	pp->src_off        = 12;
	pp->max_dest_len   = EndGetM32(&pp->buf[4]);
	pp->flags |= PPLS;

	/* search for overlay */
	pp->overlay_off = 0;
	if (EndGetM32(&pp->buf[pp->buf_len - 4]) == HUNK_BREAK) {
	    for (i = 0; i < pp->buf_len; i += 4) {
		if (EndGetM32(&pp->buf[i]) == HUNK_OVERLAY) {
		    pp->overlay_off = i;
		    break;
		}
	    }
	    if (i >= pp->buf_len) FAIL("PPLS file is corrupt");
	}
	break;
	
    /* encrypted data file
     * 0x00 identifier 'PX20'
     * 0x04 password checksum
     * 0x06 efficiency: 0x0w0x0y0z where w=9, x=9-10, y=9-12, z=9-13
     * 0x0A packed data, always multiple of 4 bytes
     * ---- 0xDDDDDDBB; DDDDDD=decrunched length, BB=start bit (0-31)
     */
    case 0x50583230: /* PX20 */
	printf("%s: encrypted PowerPacker data file (sum=%04X)\n",
	       pp->name, EndGetM16(&pp->buf[4]));
	pp->efficiency_off = 6;
	pp->src_off        = 10;
	pp->flags         |= ENCRYPTED;
	break;
	    
    default:
	return 0;
    }

    pp->src_len = ((pp->flags & PPLS) && pp->overlay_off
		   ? pp->overlay_off
		   : pp->buf_len) - pp->src_off - 4;
    if (pp->src_len < 4)    FAIL("data file is too short");
    if (pp->src_len & 0x03) FAIL("data file has bad length");
    pp->encrypted_len = pp->src_len;
    return 1;
}

/*--- main logic for decrunching/decrypting one file -------------------------*/

/* most common "starts" to packed data stream in executables */
static unsigned int common_exe_tails[] = {
    0x001809EE, 0x1006027A, 0x1806027A, 0x0806027A, 0x001809EF, 0xC3809E7E,
    0x001809F6, 0x0006027A, 0x09140DFE, 0x31009E7E, 0x9084029B, 0x09F7FFFE
};

static int ppunpack(struct ppfile *pp, unsigned int start_key,
		    unsigned int end_key, unsigned int num_cpus)
{
    unsigned int i, key, key_step;
    struct pp_units base;
    void *(*cracker)(void *);
    const unsigned char *src_end;

    /* try to recognise the file, either an executable or data file */
    if (EndGetM32(&pp->buf[0]) == HUNK_HEADER) {
	pp->flags |= EXECUTABLE;
	if (!(recog_pp_exe(pp) || recog_qpp_exe(pp) || recog_ppp_exe(pp))) {
	    FAIL("Amiga executable not packed with PowerPacker");
	}
    }
    else {
	if (!recog_pp_data(pp)) FAIL("unrecognised data file");
    }

    /* file has been recognised, time to decrypt / decrunch it */
    src_end = &pp->buf[pp->src_off + pp->src_len];
    pp->dest_len = (src_end[0] << 16) | (src_end[1] << 8) | src_end[2];

    if (pp->flags & ENCRYPTED) {
	/* encrypted, need to crack */

	/* if bigmode is turned on, use that */
	if (pp->flags & BIGMODE) {
	    pp->encrypted_len = (pp->encrypted_len + 3) >> 2;
	    pp->encrypted_len = (pp->encrypted_len - 1) & 0xFFFF;
	    pp->encrypted_len = (pp->encrypted_len + 1) << 2;
	}

	if (start_key == end_key) {
	    /* only one key */
	    if (! decrunch(pp, start_key)) {
		printf("%s: the key %08X did not decrypt the file.\n",
		       pp->name, start_key);
	    }
	    return 1;
	}

	/* need to crack */
	if (pp->src_len == pp->encrypted_len) {
	    /* can use fast cracking */
	    cracker = (pp->flags & MASTERMODE)
		? ppcrack_fast_master
		: ppcrack_fast_normal;

	    /* Halving the keyspace from 2^32 to 2^31: PowerPacker alternates
	     * between two operations; literal and match. The FIRST operation 
	     * must be a literal, as there is no data to match yet, so the
	     * first bit must be a 0 in normal mode or 1 in mastermode */
	    base.fast.drop_mask = 1 << src_end[3];
	    base.fast.drop_key  = ((EndGetM32(src_end - 4)
				    ^ ((pp->flags & MASTERMODE) ? 0 : ~0))
				   & base.fast.drop_mask);

	    /* if start key or end key are in a drop range, move them out */
	    if ((start_key & base.fast.drop_mask) == base.fast.drop_key) {
		start_key |= base.fast.drop_mask-1;
		start_key++;
	    }
	    if ((end_key & base.fast.drop_mask) == base.fast.drop_key) {
		end_key &= ~(base.fast.drop_mask-1);
		end_key--;
	    }
	    if (start_key > end_key) FAIL("no key could decrypt the file");
	}
	else {
	    /* part of the data isn't encrypted, need slow cracking */
	    cracker = ppcrack_partial;

	    /* decrunch the unencrypted part */
	    partial_init(&base.partial.state, pp,
			 (pp->flags & MASTERMODE) ? 1 : 0);
	    if (!partial_verify(&base.partial.state,
				(pp->src_len - pp->encrypted_len) / 4, 0))
	    {
		FAIL("can't decrunch unencrypted part");
	    }
	    base.partial.num_longs = pp->encrypted_len / 4;
	}

	/* if we have more CPUs than keys, reduce the number of CPUs used */
	if (num_cpus > (end_key - start_key)) num_cpus = end_key - start_key;

	/* allocate space for threads */
	base.pp = pp;
	base.thread = (pthread_t *) malloc(num_cpus * sizeof(pthread_t));
	base.unit   = (struct pp_unit *)malloc(num_cpus*sizeof(struct pp_unit));
	if (!base.thread || !base.unit) {
	    free(base.thread);
	    free(base.unit);
	    FAIL("out of memory");
	}

	/* this is an executable, try dictionary plaintext attack */
	if (pp->flags & EXECUTABLE) {
	    for (i=0; i < sizeof(common_exe_tails)/sizeof(unsigned int); i++) {
		unsigned int skip = src_end[3];
		key = common_exe_tails[i] ^
		    ((EndGetM32(src_end - 8) << (32 - skip)) |
		     (EndGetM32(src_end - 4) >> skip));
		key = (key >> (32 - skip)) | (key << skip);

		if (key < start_key || key > end_key) continue;

		base.num_units           = 1;
		base.unit[0].base        = &base;
		base.unit[0].start_key   = key;
		base.unit[0].current_key = key;
		base.unit[0].end_key     = key;
		base.unit[0].success     = 0;
		cracker(&base.unit[0]);
		if (base.unit[0].success && !(pp->flags & ALLRESULTS)) {
		    free(base.thread);
		    free(base.unit);
		    return 1;
		}
	    }
	}

	/* start threads at different parts of the key space */
	base.num_units = num_cpus;
	key_step = (end_key - start_key) / num_cpus;
	for (i = 0, key = start_key; i < num_cpus-1; i++, key += key_step) {
	    base.unit[i].base        = &base;
	    base.unit[i].start_key   = key;
	    base.unit[i].current_key = key;
	    base.unit[i].end_key     = key + key_step - 1;
	    base.unit[i].success     = 0;
	    if (pthread_create(&base.thread[i], NULL, cracker,
			       (void *) &base.unit[i]))
	    {
		WARN("failed to create thread");
		pthread_exit(NULL);
	    }
	}

	/* use the main thread for the last unit */
	base.unit[i].base        = &base;
	base.unit[i].start_key   = key;
	base.unit[i].current_key = key;
	base.unit[i].end_key     = end_key;
	base.unit[i].success     = 0;
	cracker(&base.unit[i]);

	/* wait for all other threads to end */
	for (i = 0; i < num_cpus-1; i++) pthread_join(base.thread[i], NULL);
	free(base.thread);
	free(base.unit);
    }
    else {
	/* not encrypted, just unpack */
	decrunch(pp, 0);
    }

    /* all OK */
    return 1;
}

/* --- parse options and load files ----------------------------------------- */

enum {O_KEY = 0, O_FROM, O_TO, O_PASS, O_BIG, O_ALL, O_CPUS, O_UNK};
const char *opts[] = {"key", "from", "to", "pass", "big", "all", "cpus"};

int main(int argc, char *argv[]) {
    int i, j, keymode=0, passmode=0, rangemode=0, big_mode=0, all=0, num_cpus=1;
    unsigned int start_key = 0x00000000, end_key = 0xFFFFFFFF;

    if (argc < 2) {
	fprintf(stderr,
		"Usage: %s [options] <file(s)>\n"
		"   or  %s [options] -key <key> <file(s)>\n"
		"   or  %s [options] -pass <password> <file(s)>\n"
		"   or  %s [options] [-from <key>] [-to <key>] <file(s)>\n"
		"Options: -all       continue after finding a good key\n"
		"         -cpus <N>  use N concurrent threads\n"
		"         -big       for files encrypted with PPlib v37.3\n\n"
		"Example: %s -key 30a9f3e3 secret.pp\n",
		argv[0], argv[0], argv[0], argv[0], argv[0]);
	return 1;
    }

    /* parse options */
    for (i = 1; i < argc; i++) {
	int arg = O_UNK;
	
	if (argv[i][0] != '-') break;
	
	for (j = 0; j < O_UNK; j++) {
	    if (strcmp(&argv[i][1], opts[j]) == 0) { arg = j; break; }
	}
	
	switch (arg) {
	case O_KEY:  keymode   = 1; break;
	case O_FROM:
	case O_TO:   rangemode = 1; break;
	case O_PASS: passmode  = 1; break;
	case O_BIG:  big_mode  = 1; break;
	case O_ALL:  all       = 1; break;
	case O_CPUS:                break;
	default:
	    fprintf(stderr, "Unknown option '%s'\n", argv[i]);
	    return 1;
	}
	
	if ((keymode + passmode + rangemode) > 1) {
	    fprintf(stderr, "Only use one of -key, -pass or -from/-to\n");
	    return 1;
	}
	
	/* all options except -big and -all have a parameter */
	if (arg == O_BIG || arg == O_ALL) continue;
	
	if (!argv[++i]) {
	    fprintf(stderr, "Missing argument for -%s option.\n", opts[arg]);
	    return 1;
	}
	
	if (arg == O_PASS) {
	    start_key = end_key = ppCalcPasskey((unsigned char *) argv[i]);
	    printf("Password '%s' = checksum %04X, key %08X\n", argv[i],
		   ppCalcChecksum((unsigned char *) argv[i]), start_key);
	}
	else if (arg == O_CPUS) {
	    if (sscanf(argv[i], "%d", &j) != 1 || j < 1 || j > 1024) {
		fprintf(stderr, "'%s' is a bad number of CPUs.\n", argv[i]);
		return 1;
	    }
	    num_cpus = j;
	}
	else {
	    if (sscanf(argv[i], "%8x", &j) != 1) {
		fprintf(stderr, "Bad hex number '%s' for -%s option.\n",
			argv[i], opts[arg]);
		return 1;
	    }
	    switch (arg) {
	    case O_FROM: start_key           = j; break;
	    case O_TO:   end_key             = j; break;
	    case O_KEY:  start_key = end_key = j; break;
	    }
	}
    }

    if (start_key > end_key) {
	fprintf(stderr, "Start key (%08X) is more than end key (%08X).\n",
		start_key, end_key);
	return 1;
    }

    if (start_key == end_key) {
	printf("Using only the key %08X to decrypt files.\n", start_key);
    }
    else {
	printf("Decrypting with keys from %08X to %08X.\n", start_key, end_key);
    }

    /* parse the remaining options as files */
    for (; i < argc; i++) {
	unsigned char *data = NULL;
	FILE *fh = NULL;
	size_t filelen;

	if ((fh = fopen(argv[i], "rb")) && 
	    fseek(fh, 0, SEEK_END) == 0 &&
	    (filelen = ftell(fh)) &&
	    fseek(fh, 0, SEEK_SET) == 0 &&
	    (data = (unsigned char *) malloc(filelen)) &&
	    fread(data, 1, filelen, fh) == filelen)
	{
	    struct ppfile pp;
	    memset(&pp, 0, sizeof(pp));
	    pp.name    = argv[i];
 	    pp.buf     = data;
	    pp.buf_len = filelen;
	    pp.flags   = (big_mode ? BIGMODE    : 0) |
			 (all      ? ALLRESULTS : 0);
	    ppunpack(&pp, start_key, end_key, num_cpus);
	}
	else {
	    perror(argv[i]);
	}
	if (fh) fclose(fh);
	free(data);
    }
    return 0;
}
